/************************************************************************ *********
 * Proiectarea cu Microprocesoare
 * Laboratorul 1
 * Digital I/O
 *********************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>

#define F_CPU 16000000
#include <util/delay.h>

/************************************************************************ 
 * DEFINE-uri pentru parametrizarea modulului LCD
 ************************************************************************/
#define LcdDATA_DDR	DDRC	// Portul pe care conectam firele de date la LCD-ul
#define LcdDATA_PORT	PORTC
#define LcdDATA_PIN	PINC

#define LcdCMD_DDR	DDRC	// Portul pe care conectam firele de comenzi la LCD
#define LcdCMD_PORT	PORTC
#define LcdCMD_PIN	PINC

#define LcdD4		PC0	// Pin-ul pentru firul de date D4 de pe LCD
#define LcdD5		PC1	// Pin-ul pentru firul de date D5 de pe LCD
#define LcdD6		PC2	// Pin-ul pentru firul de date D6 de pe LCD
#define LcdD7		PC3	// Pin-ul pentru firul de date D7 de pe LCD

#define LcdRS		PC4	// Pinul pentru selectare operatie (LCD)
#define LcdRW		PC5	// Pinul pentru Read/ Write (LCD)
#define LcdE		PC6	// Pinul de Enable (LCD)

#define LCD_INSTR_4wire 	0x28 	// 4 fire date, font 5x8
#define LCD_INSTR_display 	0x0C 	// Display On, Cursor Off, Blinking Off
#define LCD_INSTR_clearDisplay  0x01 	// Clear Display
#define LCD_INSTR_returnHome	0x02 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_nextLine 	0xC0 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_gotoCGRAM	0x40	// go to Character Generator RAM

#define NUM_MAX_LENGTH		9

#define RS_HIGH()		LcdCMD_PORT |= _BV(LcdRS)
#define RS_LOW()		LcdCMD_PORT &= ~_BV(LcdRS)
 
#define RW_HIGH()		LcdCMD_PORT |= _BV(LcdRW)
#define RW_LOW()		LcdCMD_PORT &= ~_BV(LcdRW)
 
#define ENABLE()		LcdCMD_PORT |= _BV(LcdE)
#define DISABLE()		LcdCMD_PORT &= ~_BV(LcdE)

/*************
 * API LCD   *
 *************/
void LCD_init();                // Initializare modul LCD.Trebuie apelata inainte de a se face orice operatie cu LCD-ul
void LCD_writeInstruction(unsigned char _instruction);	// Trimite o instructiune catre lcd (vezi datasheet)
void LCD_writeData(unsigned char _data);		// Trimite date catre LCD pentru afisare
void LCD_write(unsigned char _byte);			// trimite un byte catre LCD la modul general (nu conteaza daca e instructiune sau date)
void LCD_waitNotBusy();					// Asteptam pana cand lcd-ul devine disponibil pt o noua comanda
void LCD_waitInstructions(unsigned char _instructions);	// Asteapta un numar de cicli de ceas.
void LCD_print(char* _msg);				// Afiseaza imformatia pe LCD (doar 1 linie, primele 16 caractere)
void LCD_print2(char* _msg1, char* _msg2);		// Afisare pe 2 lini pe LCD
void LCD_printDecimal2u(unsigned int _n);		// Afisare 2 cifre in baza 10 pe LCD
void LCD_printDecimal_u(uint32_t _n);		// Afisare numar in baza 10 pe LCD
void LCD_printHexa(unsigned int _n);			// Afisare numar in baza 16 pe LCD

/******************************************************************************
 * DEFINE-uri pentru parametrizarea modulului de masurare a perioadei impulsului
 ******************************************************************************/
#define ICP1	PD6	// Portul pe care primim semnalul digital (trigger)

#define BASE_FREQUENCY	F_CPU
#define NUM_NOTES	6
#define CENTER		3
#define PERC1		0.01
#define PERC2		0.04
#define PERC3		0.08

#define E_STRING	82.4069				// Ultima coarda (6)
#define A_STRING	110.000
#define D_STRING	146.832
#define G_STRING	195.998
#define B_STRING	246.942
#define EH_STRING	329.628				// Prima coarda (1)

#define ABS(x) ((x) < 0 ? (-x) : x)

// Numarul de tacturi de ceas ale notelor
const uint32_t Center_Count[] =
{
	BASE_FREQUENCY / EH_STRING,			// High E
	BASE_FREQUENCY / B_STRING,			// B
	BASE_FREQUENCY / G_STRING,			// G
	BASE_FREQUENCY / D_STRING,			// D
	BASE_FREQUENCY / A_STRING,			// A
	BASE_FREQUENCY / E_STRING			// Low E
};

// Jumatatea distantei dintre doua note succesive in numar de tacturi
const uint32_t Transition_Count[] =
{
	BASE_FREQUENCY / ((EH_STRING + B_STRING) / 2),	// E to B
	BASE_FREQUENCY / ((B_STRING + G_STRING) / 2),	// B to G
	BASE_FREQUENCY / ((G_STRING + D_STRING) / 2),	// G to D
	BASE_FREQUENCY / ((D_STRING + A_STRING) / 2),	// D to A
	BASE_FREQUENCY / ((A_STRING + E_STRING) / 2)	// A to E
};

// Numarul de frecvente identice
//const char Note_Counts[] = {

// Simbolurile notelor
const char Note_Letter[] = {'e', 'B', 'G', 'D', 'A', 'E'};

// Frecventele notelor (pentru afisare)
const char *Note_Freq[] = {"329.62", "246.94", "195.99",
				"146.83", "110.00", "82.406"};
				
// Simbolurile pentru distanta notelor
const char *Center_Distance[] =
{
	" >",
	"   >",
	"     >",
	"       \377\377",
	"          <",
	"            <",
	"              <",
	"                "	// Clear line
};

/*************
 * API ICP
 ************/
uint16_t ov_counter;	// Contor de depasire
uint16_t rising_start;	// Timpul ultimului front crescator
uint16_t rising_stop;	// Timpul frontului crescator curent
uint8_t parity;		// Retine paritatea numarului de fronturi crescatoare

uint8_t	refresh_lcd = 0;	// Trebuie actualizat lcd-ul
uint8_t	note_index;		// Indexul notei
uint32_t counts_isr = 0;	// Numarul total de tacturi de la intrerupere
uint32_t counts_n = 0;		// Numarul de masuratori
uint32_t counts;		// Numarul de tacturi
uint32_t counts_last = 0;	// Ultimul numar de tacturi
uint32_t counts_sum = 0;	// Suma numarului de tacturi
uint16_t counts_same = 0;	// Numarul de tacturi apropiate
uint32_t diff;			// Diferenta de tacturi fata de nota centrala

// Subrutina de tratare a depasirii Timer1
ISR(TIMER1_OVF_vect) {
	ov_counter++;
}

// Subrutina de tratare a intreruperii de captura a Timer1
ISR(TIMER1_CAPT_vect) {
	// Verifica paritatea numarului frontului crescator curent
	if (parity == 0) {
		// Schimba paritatea
		parity++;
		// Salveaza timpul de start
		rising_start = ICR1;
		// Reseteaza contorul de depasire
		ov_counter = 0;
	}
	else {
		// Schimba paritatea
		parity--;
		
		// Salveaza timpul de sfarsit
		rising_stop = ICR1;
		
		// Calculeaza numarul total de tacturi
		counts_isr += (uint32_t) rising_stop - (uint32_t) rising_start +
				((uint32_t) ov_counter * (uint32_t) UINT16_MAX);
		counts_n++;
		
		// Actualizeaza lcd-ul
		if (!refresh_lcd) {
			counts = counts_isr / counts_n;
			counts_isr = 0;
			counts_n = 0;

			refresh_lcd = 1;
		}
	}
}

void find_and_write_note() {
	// Verifica daca trebuie sa calculam noua nota
	if (!refresh_lcd) {
		return;
	}

	// Determina diferenta dintre sunete
	if (counts_last > counts) {
		diff = counts_last - counts;
	}
	else {
		diff = counts - counts_last;
	}

	// Numara repetarea aceleiasi frecvente
	if (diff < Center_Count[NUM_NOTES - 1] * PERC1 && counts_same > 0) {
		counts_sum += counts;
		counts_same++;
	}
	else {
		counts_last = counts;
		counts_sum = counts;
		counts_same = 1;

		// Determina nota
		for (note_index = 0; note_index < NUM_NOTES - 1; note_index++) {
			if (counts < Transition_Count[note_index]) {
				break;
			}
		}
	}

	LCD_writeInstruction(LCD_INSTR_returnHome);
	LCD_printDecimal_u(counts_same);

	// Verifica stabilitatea sunetului
	if (counts_same < (1 << 2)) {
		refresh_lcd = 0;

		return;
	}
	counts = counts_sum / counts_same;
	counts_same = 0;

	// Reseteaza a doua linie a LCD-ului
	LCD_writeInstruction(LCD_INSTR_returnHome);
	LCD_writeInstruction(LCD_INSTR_nextLine);
	LCD_print((char *) Center_Distance[CENTER + 4]);
	LCD_writeInstruction(LCD_INSTR_returnHome);

	// Afiseaza nota si frecventa curenta (prima linie)
	LCD_writeData(' ');
	LCD_writeData(Note_Letter[note_index]);
	//LCD_writeData(' ');
	LCD_print((char *) Note_Freq[note_index]);
	LCD_writeData(' ');
	LCD_printDecimal_u(BASE_FREQUENCY / counts);	// Frecventa
	LCD_writeData('.');
	LCD_printDecimal_u(((BASE_FREQUENCY % counts) * 1000) / counts);

	// Afiseaza decalajul fata de nota (a doua linie)
	LCD_writeInstruction(LCD_INSTR_nextLine);
	if (counts < Center_Count[note_index]) {
		diff = Center_Count[note_index] - counts;
		counts = Center_Count[NUM_NOTES - 1];

		if (diff > counts * PERC3) {
			LCD_print((char *) Center_Distance[CENTER + 3]);
		}
		else if (diff > counts * PERC2) {
			LCD_print((char *) Center_Distance[CENTER + 2]);
		}
		else if (diff > counts * PERC1) {
			LCD_print((char *) Center_Distance[CENTER + 1]);
		}
		else {
			LCD_print((char *) Center_Distance[CENTER]);
		}
	}
	else if (counts > Center_Count[note_index]) {
		diff = counts - Center_Count[note_index];
		counts = Center_Count[NUM_NOTES - 1];

		if (diff > counts * PERC3) {
			LCD_print((char *) Center_Distance[CENTER - 3]);
		}
		else if (diff > counts * PERC2) {
			LCD_print((char *) Center_Distance[CENTER - 2]);
		}
		else if (diff > counts * PERC1) {
			LCD_print((char *) Center_Distance[CENTER - 1]);
		}
		else {
			LCD_print((char *) Center_Distance[CENTER]);
		}
	}

	refresh_lcd = 0;
}
void delay() {
	int i;
	
	for (i = 1; i <= 10; i++)
		_delay_ms(256);
}
 
 
int main()
{
	// Initializeaza LCD-ul
	LCD_init();
	LCD_print2("  Guitar Tuner", "      by PheonyX");
	
	// Stinge ledul
	DDRD |= _BV(7);
	PORTD &= ~_BV(7);

	// Activeaza intreruperile captura a intrarii si depasire
	TIMSK = _BV(TICIE1) | _BV(TOIE1);

	// Micsorarea zgomotului, front crescator, fara prescaler
	TCCR1B = _BV(ICNC1) | _BV(ICES1) | _BV(CS10);

	// Activeaza intreruperile
	sei();

	// Bucleaza (se ocupa timerul si interuperile)
	while(1) {
		// Actualizeaza lcd-ul
		find_and_write_note();
	}

	return 0;
}

/*****************************
 * IMPLEMENTARE API LCD.     *
 *****************************/

void LCD_reset()
{
	LcdDATA_DDR |=  (1<<LcdD4)|(1<<LcdD5)|(1<<LcdD6)|(1<<LcdD7);	// setam pinii de date ca pini de iesire
	LcdCMD_DDR  |=  (1<<LcdRS)|(1<<LcdRW)|(1<<LcdE);				// setam pinii de comenzi ca pini de iesire
	
	LcdCMD_PORT |= _BV(LcdRS)|_BV(LcdRW)|_BV(LcdE)|_BV(LcdD7)|_BV(LcdD6)|_BV(LcdD5)|_BV(LcdD4);

	_delay_ms(20);

	LcdCMD_PORT   &= ~(1<<LcdRS);									// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);									// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |= (1<<LcdD4)|(1<<LcdD5);
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	_delay_us(600);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );				// Setam linia E(nable) pe low; transferul s-a terminat

	_delay_ms(10);

	LcdCMD_PORT   &= ~(1<<LcdRS);									// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);									// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |= (1<<LcdD4)|(1<<LcdD5);
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	_delay_us(600);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );				// Setam linia E(nable) pe low; transferul s-a term

	_delay_ms(1);

	LcdCMD_PORT   &= ~(1<<LcdRS);									// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);									// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |= (1<<LcdD4)|(1<<LcdD5);
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	_delay_us(600);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );				// Setam linia E(nable) pe low; transferul s-a term

	_delay_ms(1);

	LcdCMD_PORT   &= ~(1<<LcdRS);									// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);									// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD4)&~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |=  (1<<LcdD5);									// setam pinii de comenzi ca pini de iesire
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	_delay_us(600);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );				// Setam linia E(nable) pe low; transferul s-a terminat

	_delay_ms(1);

}

void LCD_init()
{
	LCD_reset();

	LcdDATA_DDR |=  (1<<LcdD4)|(1<<LcdD5)|(1<<LcdD6)|(1<<LcdD7);	// setam pinii de date ca pini de iesire
	LcdCMD_DDR  |=  (1<<LcdRS)|(1<<LcdRW)|(1<<LcdE);				// setam pinii de comenzi ca pini de iesire

	LCD_writeInstruction(0x28); 		// Incarcam comanda: 4 bit data, 2 lines, 5x8 font
	LCD_writeInstruction(0x0C); 		// Display On, Cursor Off, Blinking Off
	LCD_writeInstruction(0x01);		// Clear Display

	LCD_writeInstruction(0x06);		// Increment, no shift
	
	LCD_writeInstruction(LCD_INSTR_returnHome);
}

void LCD_writeInstruction(unsigned char _instruction)
{
	/* TODO */					
	// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LCD_waitNotBusy();	
	
	// setam pinul RS pe low (low=instructiuni, high=date)
	RS_LOW();
	
	// setam pinul RW pe low (suntem in modul de comenzi acum)
	RW_LOW();
	
	// apelam procedura ce trimite byte-ul pe firele de date
	LCD_write(_instruction);
}


void LCD_writeData(unsigned char _data)
{
        /* TODO */
	// similar cu LCD_writeInstruction
		
	// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LCD_waitNotBusy();
	
	// setam pinul RS pe low (low=instructiuni, high=date)
	RS_HIGH();
	
	// setam pinul RW pe low (suntem in modul de comenzi acum)
	RW_LOW();
	
	// apelam procedura ce trimite byte-ul pe firele de date
	LCD_write(_data);
}

void LCD_write(unsigned char _byte)
{
	/* TODO */
	// scriem cei mai semnificativi 4 biti pe firele de date
	unsigned char aux = _byte & 0xf0;
	aux >>= 4;
	LcdDATA_DDR = 0xff;
	LcdDATA_PORT &= 0b11110000;
	LcdDATA_PORT |= aux;
	
	LcdCMD_DDR = 0xff;
	// Setam Pinul E pe high
	ENABLE();
	
	// Asteptam o perioada de timp T
	_delay_ms(5);
	
	// Setam Pinul E pe low
	DISABLE();
	
	// Asteptam o perioada de timp T
	_delay_ms(5);
	
	// scriem cei mai putin semnificativi 4 biti pe firele de date
	aux = _byte & 0x0f;
	LcdDATA_DDR = 0xff;
	LcdDATA_PORT &= 0b11110000;
	LcdDATA_PORT |= aux;
	
	// Setam Pinul E pe high
	ENABLE();
	
	// Asteptam o perioada de timp T
	_delay_ms(5);
	
	// Setam Pinul E pe low
	DISABLE();
}
/*
void LCD_waitNotBusy()
{
	unsigned char _loop = 1;

	while (_loop)
	{
		LcdDATA_DDR &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7);	// Setam pinii de date de la LCD pe in pt a citi busy flag
		LcdDATA_PORT &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); 	// Dezactivam pullup resistor pentru pinii de in
		
		LcdCMD_PORT &= ~(1<<LcdE);					// Setam pin-ul e pe low; ar trebui sa fie deja low, doar ne asiguram
		LcdCMD_PORT &= ~(1<<LcdRS);					// Setam pinul RS pe low
		LcdCMD_PORT |=  (1<<LcdRW);					// Setam pinul RW pe high (acum suntem in modul de interogare busy/adr)
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		//LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_delay_ms(1);
		_loop = LcdDATA_PIN & (1<<LcdD7);				// Citim busy flag-ul
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

		//LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_delay_ms(1);

		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		//LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_delay_ms(1);
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

		LcdDATA_DDR |= (1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); // Setam Portul de LCD ca port de iesire la loc
	}
}
*/


void LCD_waitNotBusy()
{
	unsigned char i,j;
    for(i=0;i<50;i++)        //A simple for loop for delay
		for(j=0;j<255;j++);
}

void LCD_waitInstructions(unsigned char _instructions)
{
	volatile int instr = _instructions;
	while (instr--)
		;
}


void LCD_printDecimal2u(unsigned int _n)
{
	unsigned char tmp=0;
	
	// Extragem sutele
	while(_n>=100)
		_n-=100;

	while(_n>=10){
		tmp++;
		_n-=10;
	}

	LCD_writeData(tmp+'0');
	LCD_writeData(_n+'0');
}

void LCD_printDecimal_u(uint32_t _n)
{
	char number[NUM_MAX_LENGTH];
	int i;
	
	for (i = 0; _n > 0; i++) {
		number[i] = _n % 10;
		_n /= 10;
	}

	for (i--; i >= 0; i--) {
		LCD_writeData(number[i] + '0');
	}
}

void LCD_printHexa(unsigned int _n)
{
	unsigned char _tmp = _n>>4;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
	_tmp = _n & 0x0F;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
}


void LCD_print(char* _msg)
{
	unsigned char i=0;
	for( ; _msg[i]!=0 && i<16; i++)
		LCD_writeData( _msg[i] );
}

void LCD_print2(char* _msg1, char* _msg2)
{
	LCD_writeInstruction(LCD_INSTR_clearDisplay);
	LCD_print(_msg1);
	LCD_writeInstruction(LCD_INSTR_nextLine);
	LCD_print(_msg2);
}
